newSound = love.audio.newSource
setVolume = love.audio.setVolume
getVolume = love.audio.getVolume
play = love.audio.play
getPosition = love.mouse.getPosition
setColor = love.graphics.setColor
setFont = function(font)
   if type(font)~='userdata' or tostring(font)~='Font' then
      error('This setFont is modified to accept only already-loaded fonts')
   end
   love.graphics.setFont(font)
end
newFont = love.graphics.newFont
getWidth = love.graphics.getWidth
getHeight = love.graphics.getHeight
setMode = love.graphics.setMode
rectangle = love.graphics.rectangle
text = love.graphics.printf
disp = love.graphics.print
sleep = love.timer.sleep
max = math.max

---- Callbacks

function love.keypressed(key)
   if key=='s' then setVolume(1-getVolume())
   elseif key=='q' then setVolume(0); setMode(1,1);   
   elseif key=='r' then setVolume(1); setMode(800,600)
   end
end
   
function love.mousepressed(x,y,button)
   if button~='l' then return end
   local t = which(tile,x,y)
   if t and not active_tile and not t.frozen then 
      activate(t); dragging = true 
      return
   end
   local p = which(slot,x,y)
   if p and active_tile then 
      snap(active_tile,p); dragging=true
      return 
   end
end

function love.mousereleased(x,y,button)
   if button~='l' then return end
   dragging = false
   drop(which(slot,getPosition())) 
end

function love.update(dt)
   if won or lost then love.mousepressed=nil; love.mousereleased=nil end
   if won and effect(invert,colors,dt,0.25,17.25) then play(win_sound)
   elseif lost and effect(fade,colors,dt,0.078,19) then play(lose_sound)
   elseif active_tile and dragging then
      local x,y = getPosition()
      active_tile.x, active_tile.y = x-size/2, y-size/2      
   end
end 

function love.quit()
   if not (won or lost) then play(oops_sound); sleep(1240) end
end

function love.draw()
   board:draw()
   for k=1,#tile do tile[k]:draw() end
   if active_tile then active_tile:draw() end
   score:draw()
   if won then credits() end
end
draw = love.draw

function love.load()
-- sounds
   ok_sound = newSound "ok.ogg"
   oops_sound = newSound "oops.ogg"
   win_sound = newSound "win.ogg"
   lose_sound = newSound "lose.ogg"
-- sizes and fonts
   size = 108
   local fontsize, border = 72, 10
   large = newFont(fontsize)
   small = newFont(fontsize/2)
   tiny = newFont(fontsize/4)
   local margin = size+border-fontsize
   local s = 4*size+5*border
-- colors
   white = {255,255,255}
   board_bg = {128,255,128,128}
   tile_fg = {120,48,120}
   tile_bg = {224,208,128}
   score_fg = {255,255,0}
   score_bg = {128,0,0}
   freeze_fg = tile_fg
   freeze_bg = {255,240,160}  
   colors = { tile_fg, tile_bg, board_bg }
-- the board
   board = { x=getWidth()-border-s, y=getHeight()-border-s, sx=s, 
             bg=board_bg, draw = draw_rect }
   slot = {}
   for k=0,15 do   
      local i,j = k%4, (k-k%4)/4
      slot[k+1] = { x = board.x + (j+1)*border + j*size,
                   y = board.y + (i+1)*border + i*size,
                   row=i+1, col=j+1, label=('(%i,%i)'):format(i+1,j+1) }
   end
   for k=0,4 do 
      local i,j = k,0
      slot[k+17] = { x = (j+1)*border + j*size, y = (i+1)*border + i*size,
         label='R'..k+1 }
   end 
-- the pieces
   ngrid=16; outside=5
   tile = {}
   local start = {1,17,9,13,2,6,10,18,19,7,11,15,4,20,21,16} 
   for k=1,ngrid do 
      local r=0.12
      if k<10 then r=0.65 end
      local j=start[k] 
      local x,y = slot[j].x,slot[j].y
      tile[k] = { x=x, y=y, dx=r*margin, dy=0.25*margin, sx=size,  
         fg=tile_fg, bg=tile_bg, label=k, draw = draw_tile, font = large }
      tile[k].slot = slot[j]
      slot[j].tile = tile[k]
   end
-- scorebox
   score = { x=board.x; y=border; draw = draw_tile; sx=size; sy=size/2;
      fg=score_fg; bg=score_bg; label=100; dx=border; dy=border/2; 
      font = small }
end

---- tile functions

function activate(t)
   t.x0, t.y0 = t.x, t.y
   t.fg, t.bg = t.bg, t.fg  
   active_tile = t
   end

function inactivate(t)
   t.fg, t.bg = t.bg, t.fg    
   active_tile = nil
   end

function snap(t,p) 
   if p then t.x, t.y = p.x, p.y else t.x, t.y = t.x0, t.y0 end
   end

function drop(p)
   local t = active_tile
   if t==nil or p==nil then return end
   inactivate(t) 
   if t==p.tile then snap(t); return end
   score.label = max(0,score.label-1)
   if legal(t,p) then
      if t.slot.row==nil then outside = outside-1 end
      snap(t,p)
      t.slot.tile = nil
      t.slot, p.tile = p, t 
      if outside==0 then won=true else play(ok_sound) end 
      return
   end
   if not lost then snap(t); play(oops_sound)
   else score.label=0; snap(t,p)
   end
end

function which(tile,x,y)
   for k=1,#tile do
      local t=tile[k] 
      if x>t.x and x<t.x+size and y>t.y and y<t.y+size then
         return t
      end
   end
end

function draw_tile(tile)
   draw_rect(tile); setColor(tile.fg); setFont(tile.font)
   text(tile.label, tile.x+tile.dx, tile.y+tile.dy, tile.sx)
end

function get_tile(i,j) return slot[i+4*(j-1)].tile end

function freeze(t)
   if t==nil then return end
   t.fg = freeze_fg; t.bg = freeze_bg; t.frozen=true
end

---- color functions

function invert(self) for i=1,3 do self[i]=255-self[i] end end

function fade(self) for i=1,3 do self[i]=max(0,self[i]-1) end end

function effect(func,colors,dt,bigdt,total)
   if delay then 
      delay=delay+dt; 
      if delay-delay0>bigdt then
         delay0 = delay
         for k,v in pairs(colors) do func(v) end
      end
      if delay>total then love.event.push 'q' end
   else delay=0; delay0=0; return true
   end
end

function draw_rect(r)
   setColor(r.bg); rectangle('fill', r.x, r.y, r.sx, r.sy or r.sx)
end

function credits()
   topline = (topline or 600) - 1
   setFont(tiny)
   setColor(white)
   disp ([[
        RHINO NEXT YEAR

Game engine:      LOVE 2D 
   <http://love2d.org>

          Sounds

OK move:          Neddicky
Illegal move:     Black Bustard
Losing move:      Hadeda Ibis
Win:              Chorister Robin

Recorded in South Africa

Sound engineer: Charles Hesse
<http://www.tropicalbirding.com/>
Repository: Internet Bird Collection
<http://ibc.lynxeds.com>


Author:           Dirk Laurie
Date:             (c) 2011
License:          zlib/libpng
]],50,topline)
end


--- rules of the game

love.filesystem.load("legal.luac")()

